; This is the function that is actually called from AMOEBA.
; Since only the P array can be passed from amoeba, other variables have to
; be available by other means and are read from the COMMON block in cb_sfh.jou
;
; The difference between model and data arrays is calculated within a number
; of boxes which can have different weights and sample different parts of 
; the Hess diagram. Various options are available to define which quantity
; is actually minimized.

function amo_sfh, P

; P is the SFR in each bin and for each isochrone set (metallicity)
; It should have [n_elements(cb_bins)-1] * cb_nz elements

@cb_sfh.jou

  cb_psave = P
  cb_nit = cb_nit+1

  print, 'AMO_SFH: Niter= ', cb_nit

; Convert logarithmic P values in each bin to real SFRs for each
; isochrone

  nloga = n_elements(cb_loga)
  sfr = fltarr(nloga*cb_nz)

  k = 0
  for j=0,cb_nz-1 do begin
    for i=0,n_elements(cb_bins)-2 do begin
      w = where(cb_loga ge cb_bins[i] AND cb_loga lt cb_bins[i+1])
      if (w[0] gt -1) then sfr[w+j*nloga] = 10^(P[k])
      print, '  Iso = ',strn(j+1),'/',strn(cb_nz), $
             '  LogAge = ',strn(cb_bins[i]),'..',strn(cb_bins[i+1]), $
	     '  Log(SFR) = ',strn(P[k])
      k = k + 1
    end
  end

; Scale the individual isochrones by the SFR values and co-add the
; Hess diagrams

  sfhcmd, cb_loga, cb_ipixels, sfr=sfr, pixels=mpixels, $
    cpixels=cb_cpixels, cmode=cb_cmode

; We use the 25 percentile of the non-zero pixel value distribution
; as the normalisation point

  if (cb_p25 lt 0) then begin
    wpos = where(cb_dpixels gt 0)
    dpix_pos = cb_dpixels[wpos]
    s = sort(dpix_pos)
    ns = n_elements(s)
    cb_p25 = dpix_pos(s[ns*0.25])
    print, 'P25 = ',cb_p25
  end

  p25 = cb_p25

; A bit of array manipulation to get the model pixels as a true 2-D array 
; instead of a 3-D array with the first dimension equal to 1

  sz = size(mpixels)
  pdx = sz[2]
  pdy = sz[3]
  mp2d = fltarr(pdx, pdy)
  mp2d(*,*) = mpixels(0,*,*)
  boxes = cb_boxes

; Now calculate the model vs data difference in each box

  dtot = 0.

  for i=0,n_elements(boxes(0,*))-1 do begin
    x1 = fix(pdx * (boxes[0,i]-cb_xr[0])/(cb_xr[1]-cb_xr[0]))
    x2 = fix(pdx * (boxes[1,i]-cb_xr[0])/(cb_xr[1]-cb_xr[0]))
    y1 = fix(pdy * (boxes[2,i]-cb_yr[0])/(cb_yr[1]-cb_yr[0]))
    y2 = fix(pdy * (boxes[3,i]-cb_yr[0])/(cb_yr[1]-cb_yr[0]))
    if (x1 gt x2) then begin
      xtmp = x1
      x1 = x2
      x2 = xtmp
    end
    if (y1 gt y2) then begin
      ytmp = y1
      y1 = y2
      y2 = ytmp
    end

    if (x1 lt 0) then x1 = 0
    if (x2 gt pdx-1) then x2 = pdx-1
    if (y1 lt 0) then y1 = 0
    if (y2 gt pdy-1) then y2 = pdy-1

    mptmp = mp2d[x1:x2,y1:y2]
    dptmp = cb_dpixels[x1:x2,y1:y2]

    case cb_bmode of
      1: begin  ; Squared logarithmic difference, pixel by pixel
           dtmp = total( (alog(mptmp+p25) - alog(dptmp+p25) )^2. ) * boxes[4,i]
	 end
      2: begin  ; Squared difference, pixel by pixel
           dtmp = total( (mptmp - dptmp)^2. ) * boxes[4,i]
         end
      3: begin  ; Squared difference, box by box
           dtmp = (total(mptmp) - total(dptmp))^2. * boxes[4,i]
         end
      4: begin  ; As (2), but normalised to data pixel values
           dtmp = total( ((mptmp - dptmp) / (dptmp+p25))^2. ) * boxes[4,i]
         end
      5: begin  ; As (3), but normalised to data box values
           dtmp = ((total(mptmp) - total(dptmp)) / total(dptmp+p25))^2. * boxes[4,i]
         end
      6: begin  ; Difference between luminosity functions in each box
           sztmp = size(mptmp)
	   dtmp = 0.
	   for y=0,sztmp[2]-1 do $
	     dtmp = dtmp + (total(mptmp[*,y]) - total(dptmp[*,y]))^2. 
	   dtmp = dtmp * boxes[4,i]
         end
      7: begin  ; Maximum likelihood with Poissonian statistics
           pmin = 1d-20
	   dtmp = 0.
           for j=0,n_elements(mptmp)-1 do begin
	     mu = mptmp[j]
	     k  = dptmp[j]
	     if (mu lt 10 AND k lt 10) then $
	       prob = ppoisson(1d*k, 1d*mu) $
             else $
	       prob = pgauss(1d*k, 1d*mu, sqrt(mu))
;             print, prob, alog(prob), dtmp
	     if (prob lt pmin) then prob = pmin
             dtmp = dtmp - alog(prob)
;             print, k, mu, prob, alog(prob), dtmp
	   end
	   dtmp = dtmp * boxes[4,i]
         end
      8: begin  ; Maximum likelihood to LF with Poissonian statistics
           pmin = 1d-20
	   dtmp = 0.
	   sztmp = size(mptmp)
           for y=0,sztmp[2]-1 do begin
	     mu = total(mptmp[*,y])
	     k  = total(dptmp[*,y])
	     if (mu lt 10 AND k lt 10) then $
	       prob = ppoisson(1d*k, 1d*mu) $
             else $
	       prob = pgauss(1d*k, 1d*mu, sqrt(mu))
	     if (prob lt pmin) then prob = pmin
             dtmp = dtmp - alog(prob)
	   end
	   dtmp = dtmp * boxes[4,i]
         end
    end

    print, '  Box '+strn(i+1)+': ['+strn(boxes[0,i])+'..'+strn(boxes[1,i])+','+$
                               strn(boxes[2,i])+'..'+strn(boxes[3,i])+$
           ']  Diff='+strn(dtmp)

    dtot = dtot + dtmp

;    print, boxes(*,i)
;    print, sz, pdx, pdy
;    print, x1, x2, y1, y2
  end

  print, '  Total Diff='+strn(dtot)

  cb_mpixels = mpixels
;  return, diff
  return, dtot

end
