; Program to derive a star formation history from an observed colour-magnitude
; diagram. 
; A library of stellar isochrones is read from 'isofile' and populated
; according to an assumed IMF. A number of additional steps are taken to 
; ensure a more realistic approximation to real data:
;
; 1) Random errors (provided by the user as a function of magnitude) are
;    added to the isochrone values.
; 2) A range of reddenings (specificed by the user) are added to the colours
; 3) Observational completeness functions are applied
; 4) Some provision for adding binaries is included
;
; The AMOEBA routine is used to minimize the difference between observed and
; simulated Hess diagrams as a function of the SFH. 
;
; This software comes with no warranty whatsoever. Most likely it contains
; many bugs, some of which could seriously compromise your science.
; Use it at your own risk.
; 
; Soeren S. Larsen
; S.S.Larsen@uu.nl
; 20.05.2010
;
; s.larsen@astro.ru.nl

pro wcskwd, hdr, xr, yr, pixels
  sz = size(pixels)
  sxaddpar, hdr, 'WCSAXES', 2
  sxaddpar, hdr, 'CRPIX1', 0
  sxaddpar, hdr, 'CRPIX2', 0
  sxaddpar, hdr, 'CRVAL1', xr[0]
  sxaddpar, hdr, 'CRVAL2', yr[0]
  sxaddpar, hdr, 'CDELT1', (xr[1]-xr[0])/sz[1]
  sxaddpar, hdr, 'CDELT2', (yr[1]-yr[0])/sz[2]
  sxaddpar, hdr, 'CTYPE1', 'MAG'
  sxaddpar, hdr, 'CTYPE2', 'MAG'
end

; -------------------------------------------------------------------------

pro fitsfh, datam1, datam2,  $
; These parameters can only be specified on the command line (many of them
; are long arrays
  loga_iso=loga_iso, $
  errm1=errm1, errm2=errm2, $
  save_pixels=save_pixels, load_pixels=load_pixels, $
  cmpl=cmpl, $             
  noshow=noshow, $         
  pfile=pfile, $           ; Name of parameter file
; The following parameters can also be specified in the parameter file
  xr=xr, yr=yr, ymag=ymag, $
  mrng=mrng, $
  ktype=ktype, kwidth=kwidth, hdim=hdim, $
  ab=ab, nab=nab, am_ab=am_ab, aabfile=aabfile, $
  isofile=isofile, acol=acol, mcol=mcol, magcol=magcol, bins=bins, $
  niter=niter, $
  logfile=logfile, boxes=boxes, bmode=bmode, cmode=cmode, $
  amoeba_itmax=amoeba_itmax, $
  fbin=fbin, dbin=dbin, nqsub=nqsub, qmin=qmin, qmax=qmax

  VERSION = '0.51d'

; Parameters:
; datam1 = Observed magnitudes in filter1
; datam2 = Observed magnitudes in filter2
; xr     = Range of datam1-datam2 values to consider
; yr     = Range of datam[ymag] values to consider
; ktype  = Kernel type for Hess diagrams (SQUARE / GAUSS / DISC / DELTA)
; kwidth = Width of kernel in same units as x-axis
;            SQUARE: kwidth = dimension of the kernel
;            DISC  : kwidth = radius of disc
;            GAUSS : kwidth = sigma of Gaussian
; hdim   = Dimensions of Hess diagram
; ymag   = filter to plot on y-axis (1=filter1, 2=filter2). E.g.
;          B-V vs V: datam1=B, datam2=V, ymag=1
;          V-I vs V: datam1=V, datam2=I, ymag=0
; mrng   = Mass range for normalisation of SFR
; loga_iso = log(age) values to read from isochrone library.
;          The array need not be an exact match - the closest match to each
;          log(age) value will be read
; ab     = B-band extinction values to add. This can be a single value,
;          or a range ([AB1,AB2])
; aabfile = text file with log(age) in first column and AB1, AB2 values in
;          second and third column. This allows use of an age-dependent
;          extinction correction
; nab    = Number of discrete ab values between AB1 and AB2 to use
; am_ab  = [A(filter1)/AB, A(filter2)/AB]
; bins   = Array specifying the age bins to fit. Isochrones within each
;          bin will be "locked" together
; isofile = Text file(s) containing the isochrone data.
;          This can be an array of several file names, for example for
;          different metallicities.
; acol    = Column in isochrone file with log(age)
; mcol    = Column in isochrone file with (initial) mass
; magcol  = Columns in isochrone file with magnitudes
; errm1  = Array of errors in filter1:
;           errm1(0,*) = mag1, errm1(1,*) = err(mag1)
; errm2  = Array of errors in filter2
; save_pixels = name of text file where the Hess diagrams of each isochrone
;          will be stored. These can then be read in for a subsequent run
;          where the parameters affecting the Hess diagrams are unchanged.
;          This is usually much faster than generating them.
; load_pixels = name of text file containing Hess diagrams of the isochrones.
; niter  = Number of iterations, changing the initial guesses a bit for
;          each one.
; logfile = Logfile containing information about the fits (and the resulting
;          SFH). Can be set to 'NONE' if no logfile is desired.
; boxes  = name of text file containing list of rectangular regions in the
;          Hess diagram to fit and their weights.
; bmode  = Flag to indicate how the fit is done (see amo_sfh.pro)
;          Recommended and default value = 7  (maximum likelihood fit)
; cmpl   = Completeness data:
;          cmpl(0,*) = mag1
;          cmpl(1,*) = mag2
;          cmpl(2,*) = completeness (number between 0 and 1)
; cmode  : multiply by completeness array *before* or *after* 
;          smoothing by the error function (BEFORE / AFTER)
; amoeba_itmax : Maximum number of iterations in AMOEBA
; fbin  : Fraction of stars which have an unresolved companion
; dbin  : Distribution of binary companion masses (DELTA/IMF/UNIFORM))
; nqsub : Number of sub-bins for evaluation companion MF
; qmin,qmax  : Minimum,maximum mass ratio to consider for binaries

@cb_sfh.jou

  if (n_elements(pfile) eq 0) then pfile='fitsfh.par'
  if (n_elements(xr) eq 0) then xr = getpar(pfile,'xr',default='-1 3')*1.0
  if (n_elements(yr) eq 0) then yr = getpar(pfile,'yr',default='-2 10')*1.0
  if (n_elements(ktype) eq 0) then ktype = strupcase(getpar(pfile,'ktype',default='GAUSS'))
  if (n_elements(kwidth) eq 0) then kwidth = getpar(pfile,'kwidth',default='0.02')*1.0
  if (n_elements(hdim) eq 0) then hdim = getpar(pfile,'hdim',default='200')*1
  if (n_elements(mrng) eq 0) then mrng = getpar(pfile,'mrng',default='0.15 100')*1.0
  if (n_elements(ymag) eq 0) then ymag = getpar(pfile,'ymag',default='0')*1.0
  if (n_elements(loga_iso) eq 0) then $
    loga_iso = 6.6 + findgen(40)/20.
  if (n_elements(bins) eq 0) then bins = getpar(pfile,'bins',default='6.5 7.0 7.5 8.0 8.5')*1.0
  if (n_elements(isofile) eq 0) then $
    isofile = getpar(pfile,'isofile',default="${HOME}/cats/iso2002/iso/isoc_z019.dat")
  if (n_elements(acol) eq 0) then acol = getpar(pfile,'acol',default='1')*1
  if (n_elements(mcol) eq 0) then mcol = getpar(pfile,'mcol',default='2')*1
  if (n_elements(magcol) eq 0) then magcol = getpar(pfile,'magcol',default='10 12')*1
  if (n_elements(niter) eq 0) then niter = getpar(pfile,'niter',default='4')*1
  if (n_elements(logfile) eq 0) then logfile = getpar(pfile,'logfile',default='sfh.log')
  if (logfile ne 'NONE') then openu,loglun,logfile,/get_lun, /append, width=1000
  if (n_elements(boxes) eq 0) then begin
    tmps = getpar(pfile,'boxes')
    if (strlen(tmps) gt 0) then boxes = tmps
  end
  if (n_elements(boxes) gt 0) then $
    cb_boxes = rdtxt(boxes,[1,2,3,4,5]) $
  else $
    cb_boxes = [xr[0], xr[1], yr[0], yr[1], 1.0]
  if (n_elements(bmode) eq 0) then bmode = getpar(pfile,'bmode',default='7')*1
  if (n_elements(cmode) eq 0) then cmode = strupcase(getpar(pfile,'cmode',default='BEFORE',allowed='BEFORE AFTER'))
  if (n_elements(amoeba_itmax) eq 0) then amoeba_itmax = getpar(pfile,'amoeba_itmax',default='2000')*1
  if (n_elements(fbin) eq 0) then fbin = getpar(pfile,'fbin',default='0')*1.0
  if (n_elements(dbin) eq 0) then dbin = strupcase(getpar(pfile,'dbin',default='DELTA'))
  if (n_elements(nqsub) eq 0) then nqsub = getpar(pfile,'nqsub',default='10')*1
  if (n_elements(qmin) eq 0) then qmin = getpar(pfile,'qmin',default='0.5')*1.0
  if (n_elements(qmax) eq 0) then qmax = getpar(pfile,'qmax',default='1.0')*1.0

  if (n_elements(aabfile) eq 0) then begin
    tmps = getpar(pfile,'aabfile')
    if (strlen(tmps) gt 0) then aabfile = tmps
  end
  if (n_elements(ab) eq 0) then ab = getpar(pfile,'AB',default='0.0')*1.0
  if (n_elements(nab) eq 0) then nab = getpar(pfile,'NAB',default='10')*1

  if (n_elements(am_ab) eq 0) then begin
    tmps = getpar(pfile,'am_ab')
    if (strlen(tmps[0]) gt 0) then am_ab = tmps*1.0
  end

  if (dbin eq 'DELTA') then nbsub = 2

  cb_bmode = bmode
  cb_cmode = cmode
  cb_nz = n_elements(isofile)

  ustr = getenv('USER')+'@'+getenv('HOST')
  pwdstr = getenv('PWD')
  dstr = systime()

  print, 'FITSFH '+VERSION

  if (logfile ne 'NONE') then begin
    printf,loglun, '# FITSFH '+VERSION
    printf,loglun, '# '+ustr
    printf,loglun, '# '+pwdstr
    printf,loglun, '# '+dstr
    printf,loglun, '# Bmode = ',bmode
    printf,loglun, '# Boxes: '
    if (n_elements(boxes) gt 0) then $
      printf,loglun, '#  Read from '+boxes
    for i=0,n_elements(cb_boxes(0,*))-1 do begin
      printf, loglun, '#  Box['+strn(i+1)+'] = ['+strn(cb_boxes[0,i])+':'+strn(cb_boxes[1,i])+','+strn(cb_boxes[2,i])+':'+strn(cb_boxes[3,i])+'], w='+strn(cb_boxes[4,i])
    end
    printf,loglun, '# Cmode = ',cmode
    printf,loglun, '# Isofile(s) = '
     for i=0,cb_nz-1 do printf, loglun, '#  ['+strn(i+1)+'] '+isofile[i]    
    printf,loglun, '# Mcol = ',mcol
    printf,loglun, '# Acol = ',acol
    printf,loglun, '# Magcol = ',magcol
    printf,loglun, '# Ymag = ',ymag
    printf,loglun, '# Mrng = ',mrng
    printf,loglun, '# AB = ',ab
    printf,loglun, '# NAB = ',nab
    if (n_elements(am_ab) gt 0) then printf,loglun, '# AM_AB = ',am_ab
    printf,loglun, '# NITER = ',niter
    if (n_elements(aabfile) gt 0) then printf,loglun, '# AABFILE = ',aabfile
    printf,loglun, '# Fbin = ',fbin
    printf,loglun, '# Dbin = ',dbin
    printf,loglun, '# Nqsub = ',nqsub
    printf,loglun, '# Qmin = ',qmin
    printf,loglun, '# Qmax = ',qmax
  end

  if (n_elements(noshow) eq 0) then begin
    !P.MULTI = [0,2,2,0,0]
    loadct,5
  end
  hess_sfh, datam1, datam2, ymag=ymag, xr=xr, yr=yr, pixels=dpixels, $
    ktype=ktype, sigker=kwidth, nx=hdim, ny=hdim, $
    logscl=0.01, title='Data', noshow=noshow
  if (n_elements(noshow) eq 0) then pboxes,cb_boxes

  if (n_elements(cmpl) gt 0) then begin
    print, 'Setting up completeness array..'
    dsz = size(dpixels)
    cpixels = cmplarr(cmpl, dsz[1], dsz[2], xr, yr, ymag=ymag)
  end else begin
    print, 'Warning: no correction for incompleteness!'
;    cpixels = 0.
    cpixels = 1.0 + (dpixels - dpixels)
  endelse

; Load extinction model from file, if applicable


  if (n_elements(aabfile) gt 0) then begin
    print, 'Loading extinction model from '+aabfile
    xx = rdtxt(aabfile,[1,2,3])
    aab1 = interpol(xx(1,*), xx(0,*), loga_iso)
    aab2 = interpol(xx(2,*), xx(0,*), loga_iso)
  end

  print, 'Loading isochrones from '
  for i=0,cb_nz-1 do print, ' ['+strn(i+1)+'] '+isofile[i]
  nloga = n_elements(loga_iso)

; This is just to set up the plot window

  isocmd, loga_iso[0], xr=xr, yr=yr, pixels=ptmp, isofile=isofile[0], $
     ktype=ktype, kwidth=kwidth, hdim=hdim, mrng=mrng, $
     acol=acol, mcol=mcol, magcol=magcol, ymag=ymag, /noshow
  sz = size(ptmp)
  ipixels = fltarr(nloga*cb_nz, sz[1], sz[2])

; Either load isochrones Hess diagrams from a previous run, or generate them
; from scratch

  if (n_elements(load_pixels) gt 0) then begin
    openr,plun,load_pixels,/get_lun
    readu,plun,ipixels
    free_lun, plun
    if (n_elements(noshow) eq 0) then begin
      for i=0,nloga*cb_nz-1 do begin
	!P.MULTI = [3,2,2,0,0]
	phess, ipixels(i,*,*), xr, yr, logscl=0.05
      end
    end
  endif else begin
    k = 0
    for j=0,cb_nz-1 do begin
      for i=0,nloga-1 do begin
	if (n_elements(noshow) eq 0) then !P.MULTI = [3,2,2,0,0]

	if (n_elements(aab1) gt 0) then $
	  iab = [aab1[i], aab2[i]] $
	else if (n_elements(ab) gt 0) then $
	  iab = ab $
	else iab = 0.0

	if (n_elements(iab) eq 2) then $
	  if (iab[0] eq iab[1]) then iab = iab[0]

	print, 'log(age) = ',strn(loga_iso[i]), ' AB = ',iab

	isocmd, loga_iso[i], xr=xr, yr=yr, pixels=ptmp, ab=iab, nab=nab, $
	  am_ab=am_ab, $
	  mrng=mrng, $
	  isofile=isofile[j], acol=acol, mcol=mcol, magcol=magcol, $
	  errm1=errm1, errm2=errm2, cpixels=cpixels, cmode=cb_cmode, $
	  ymag=ymag, $
	  ktype=ktype, kwidth=kwidth, hdim=hdim, $
	  fbin=fbin, dbin=dbin, nqsub=nqsub, qmin=qmin, qmax=qmax, $
	  noshow=noshow
	ipixels[k,*,*] = ptmp
	k = k+1
	if (n_elements(noshow) eq 0) then pboxes, cb_boxes
      end
    end
  endelse

; Save the isochrone Hess diagrams for use in another run, if desired

  if (n_elements(save_pixels) gt 0) then begin
    openw,plun,save_pixels,/get_lun
    writeu,plun,ipixels
    free_lun, plun
  end

; Set up for the SFH fit

  print, 'Setting up for SFH fit..'

  cb_loga    = loga_iso
  cb_dpixels = dpixels
  cb_ipixels = ipixels
  cb_bins    = bins
  cb_xr      = xr
  cb_yr      = yr
  cb_cpixels = cpixels

; Show the Hess diagram for continuous, uniform SFH

  sfhcmd, loga_iso, ipixels, pixels=mpixels, cpixels=cb_cpixels, cmode=cb_cmode
  if (n_elements(noshow) eq 0) then begin
    !P.MULTI = [3,2,2,0,0]
    phess, mpixels, xr, yr, logscl=0.01, title='Uniform SFR'
    pboxes, cb_boxes
  end

; Initialise for AMOEBA

  nbins = n_elements(bins)
  P0 = fltarr(cb_nz*(nbins-1)) + alog10(total(dpixels)/total(mpixels)) 
  Scl = 1.0
  P  = P0
  cb_nit = 0
  cb_p25 = -1.
  diff_min = amo_sfh(P)
  Pbest = P

  for it=1,niter do begin

    cb_nit = 0
    print, amo_sfh(P)
    cb_nit = 0
    result = amoeba(1E-5, Function_Name='amo_sfh', P0=P, Scale=Scl, nmax=amoeba_itmax)
    P = Result

    if (result[0] eq -1) then begin
      print, 'Warning: NMAX exceeded in AMOEBA! Using last solution'
      P = cb_psave
    end

    print, 'Restarting AMOEBA..'
    cb_nit = 0
    result = amoeba(1E-6, Function_Name='amo_sfh', P0=P, Scale=Scl/2, nmax=amoeba_itmax/2)
    P = result

    if (result[0] eq -1) then begin
      print, 'Warning: NMAX exceeded in AMOEBA! Using last solution'
      P = cb_psave
    end

    if (n_elements(noshow) eq 0) then begin
      !P.MULTI = [2,2,2,0,0]
      phess, cb_mpixels, xr, yr, logscl=0.01, title='Best Fit'

      !P.MULTI = [1,2,2,0,0]
      phess, dpixels-cb_mpixels, xr, yr, title='Residuals'
    end

    diff = amo_sfh(P)
    if (diff lt diff_min) then begin
      diff_min = diff
      Pbest = P
    end

    if (logfile ne 'NONE') then begin
      printf,loglun, '# It   = ',it
      printf,loglun, '# P    = ',P
      printf,loglun, '# diff = ',diff
    end

    P = P0 + (randomu(seed,n_elements(P0)) - 0.5)
  end

; Plot the Hess diagram for the best-fitting SFH

  diff = amo_sfh(Pbest)

  if (n_elements(noshow) eq 0) then begin
    !P.MULTI = [2,2,2,0,0]
    phess, cb_mpixels, xr, yr, logscl=0.01, title='Best Fit'
    pboxes, cb_boxes
  end

; And the residuals..

  if (n_elements(noshow) eq 0) then begin
    !P.MULTI = [1,2,2,0,0]
    phess, dpixels-cb_mpixels, xr, yr, title='Residuals'
  end

; Print the fit results to STDOUT and (optionally) to the log file

  print, Bins
  print, Pbest
  if (logfile ne 'NONE') then begin
    printf,loglun,'# Pbest  = ',Pbest
    printf,loglun,'# Bins    = ',bins
    printf,loglun,'#     Iso  Log(age [yr])  log(SFR [Msun/yr])'
  end

  if (logfile ne 'NONE') then begin
    for j=0,cb_nz-1 do begin
      for i=0,n_elements(bins)-2 do begin
	w = where(loga_iso ge bins[i] AND loga_iso lt bins[i+1])
	if (w[0] gt -1) then $
	  for l=0,n_elements(w)-1 do $
	    printf, loglun, j+1, loga_iso[w[l]], Pbest[i+j*(n_elements(bins)-1)]-6.
      end
    end

    printf,loglun,'# --------------------------------'
  end

; Save the Hess diagrams for each SFH bin in a multi-extension FITS file
  
  dummy = 1.0
  mkhdr, phdr, dummy, /extend
  sxaddpar, phdr, 'NAXIS', 0
  sxaddpar, phdr, 'COMMENT', 'FITSFH '+VERSION
  writefits, 'isohess.fits', dummy, phdr

  nloga = n_elements(loga_iso)
  for iz=0,cb_nz-1 do begin
    for i=0,n_elements(bins)-2 do begin
      ppixels = dpixels - dpixels
      w = where(loga_iso ge bins[i] AND loga_iso lt bins[i+1])
      if (w[0] gt -1) then begin
	for j=0,n_elements(w)-1 do begin
	  k = w[j]
	  if (k lt nloga-1) then $
	    da = 10^(loga_iso[k+1]-6.0) - 10^(loga_iso[k]-6.0)
	  ppixels = ppixels + ipixels(k+iz*nloga,*,*) * da * Pbest[i+iz*(n_elements(bins)-1)]
	end
      end
      mkhdr, ihdr, ppixels, /image
      sxaddpar, ihdr, 'COMMENT', 'Isofile = '+isofile[iz]
      sxaddpar, ihdr, 'COMMENT', 'Log(age [yr]) = '+strn(bins[i])+'..'+strn(bins[i+1])
      sxaddpar, ihdr, 'COMMENT', 'Log(SFR [Msun/yr]) = '+strn(Pbest[i+iz*(n_elements(bins)-1)]-6.)
      wcskwd, ihdr, xr, yr, ppixels
      writefits, 'isohess.fits', ppixels, ihdr, /append
    end
  end

; Save the Data, Model and Residual Hess diagrams in a multi-extension FITS file

  ; Store some info about the fit in the 0th FITS extension:

  dummy = 1.0
  sz = size(dpixels)
  mkhdr, phdr, dummy, /extend
  sxaddpar, phdr, 'NAXIS', 0
  sxaddpar, phdr, 'COMMENT', 'FITSFH '+VERSION
  sxaddpar, phdr, 'COMMENT', ustr
  sxaddpar, phdr, 'COMMENT', pwdstr
  sxaddpar, phdr, 'COMMENT', dstr
  sxaddpar, phdr, 'COMMENT', 'Bmode = '+strn(bmode)
  sxaddpar, phdr, 'COMMENT', 'Cmode = '+strn(cmode)
  sxaddpar, phdr, 'COMMENT', 'Niter = '+strn(niter)
  sxaddpar, phdr, 'COMMENT', 'Mrng  = '+strn(mrng)

  for i=0,cb_nz-1 do $
    sxaddpar, phdr, 'COMMENT', 'Isofile['+strn(i+1)+'] = '+isofile[i]
  if (n_elements(ab) gt 0) then $
    sxaddpar, phdr, 'COMMENT', 'AB = '+strn(ab)
  if (n_elements(am_ab) gt 0) then $
    sxaddpar, phdr, 'COMMENT', 'AM/AB = '+strn(am_ab)
  if (n_elements(aabfile) gt 0) then $
    sxaddpar, phdr, 'COMMENT', 'AABFILE = '+aabfile
  sxaddpar, phdr, 'COMMENT', 'Fbin = '+strn(fbin)
  if (fbin gt 0) then begin
    sxaddpar, phdr, 'COMMENT', 'Dbin = '+strn(dbin)
    sxaddpar, phdr, 'COMMENT', 'Qmin = '+strn(qmin)
    sxaddpar, phdr, 'COMMENT', 'Qmax = '+strn(qmax)
  end
  sxaddpar, phdr, 'COMMENT', 'Ktype = '+ktype
  sxaddpar, phdr, 'COMMENT', 'Kwidth = '+strn(kwidth)
  sxaddpar, phdr, 'COMMENT', 'Hdim = '+strn(hdim)


  for i=0,n_elements(cb_boxes(0,*))-1 do begin
    sxaddpar, phdr, 'COMMENT', 'Box['+strn(i+1)+'] = ['+strn(cb_boxes[0,i])+':'+strn(cb_boxes[1,i])+','+strn(cb_boxes[2,i])+':'+strn(cb_boxes[3,i])+'], w='+strn(cb_boxes[4,i])
  end

  for j=0,cb_nz-1 do begin
    for i=0,n_elements(bins)-2 do begin
      sxaddpar, phdr, 'COMMENT', 'Bin['+strn(i+1)+']: Log(age)= '+strn(bins[i])+'..'+strn(bins[i+1])
      w = where(loga_iso ge bins[i] AND loga_iso lt bins[i+1])
      if (w[0] gt -1) then $
	for k=0,n_elements(w)-1 do $
	  sxaddpar, phdr, 'COMMENT', 'Iso Log(age [yr]) Log(SFR [Msun/yr]) = '+strn(j+1)+' '+strn(loga_iso[w[k]])+' '+strn(Pbest[i+j*(n_elements(bins)-1)]-6.)
    end
  end

  writefits, 'fitsfh.fits', dummy, phdr

  ; Then write the data

  mkhdr, dhdr, dpixels, /image
  sxaddpar, dhdr, 'EXTNAME', 'DATA'
  wcskwd, dhdr, xr, yr, dpixels
  writefits, 'fitsfh.fits', dpixels, dhdr, /append

  ; and the best-fitting model

  sz = size(cb_mpixels)
  mpixels = fltarr(sz[2], sz[3])
  mpixels[*,*] = cb_mpixels[0,*,*]

  mkhdr, mhdr, mpixels, /image
  sxaddpar, mhdr, 'EXTNAME', 'MODEL'
  wcskwd, mhdr, xr, yr, mpixels
  writefits,'fitsfh.fits',mpixels, mhdr, /append

  ; and the residuals (just Data - Model)

  mkhdr, rhdr, dpixels-mpixels, /image
  sxaddpar, rhdr, 'EXTNAME', 'RESID'
  wcskwd, rhdr, xr, yr, mpixels
  writefits,'fitsfh.fits',dpixels-mpixels, rhdr, /append

  ; and the completeness array

  mkhdr, chdr, cpixels, /image
  sxaddpar, chdr, 'EXTNAME', 'CMPL'
  wcskwd, chdr, xr, yr, cpixels
  writefits,'fitsfh.fits',cpixels, chdr, /append


  if (logfile ne 'NONE') then free_lun,loglun

end
