; --------------------------------------------------------------------------

function dndm, m, imf=imf
; IMF prescription

  if (n_elements(imf) eq 0) then imf='SALPETER'

  case imf of
    'SALPETER' : begin
		   dn = m^(-2.35)
                 end
  end

  return, dn
end

; --------------------------------------------------------------------------

function dndp, m, q, dbin=dbin, imf=imf, qmin=qmin, qmax=qmax
; Binary mass ratio function
; m = primary mass
; q = secondary/primary mass ratio (1 = equal mass binary)
; dbin = Delta function  (all binaries have mass ratio = qmax)
;      = Sample from IMF
;      = Uniform weighting

  if (n_elements(dbin) eq 0) then dbin = 'DELTA'

  dn = 0.

  case dbin of
    'DELTA' : begin       
          if (q gt qmax-0.01) then dn = 1.
        end
    'IMF' : begin
          dn = dndm(q*m, imf=imf)
        end
    'UNIFORM' : begin
          dn = 1.
        end
    else : print,'** ISOCMD: unknown binary mass distribution: ',dbin
  end

  return, dn

end

; ==========================================================================

pro isocmd, logage, isofile=isofile, acol=acol, mcol=mcol, $
      magcol=magcol, am_ab=am_ab, nsub=nsub, xr=xr, yr=yr, $
      pixels=pixels, xp=xp, yp=yp, ab=ab, nab=nab, ymag=ymag, $
      logscl=logscl, noshow=noshow, errm1=errm1, errm2=errm2, $
      mrng=mrng, fbin=fbin, dbin=dbin, nqsub=nqsub, qmin=qmin, qmax=qmax, $
      cpixels=cpixels, cmode=cmode, neg=neg, charsize=charsize, $
      xtickformat=xtickformat, ytickformat=ytickformat, $
      xtitle=xtitle, ytitle=ytitle, title=title, $
      ktype=ktype, kwidth=kwidth, hdim=hdim

; Produce a colour-magnitude Hess diagram for a single isochrone.
;
; The density of the Hess diagram will be scaled to a total mass of
; 1 M(Sun) over the mass range MRNG (default 0.15..100 Msun, for a
; normalisation roughly similar to that of a Kroupa IMF).
;

  if (n_elements(isofile) eq 0) then $
    isofile = "${HOME}/cats/iso2002/iso/isoc_z008.dat"
  if (n_elements(nsub) eq 0) then nsub = 10                ; Interpolation bins
  if (n_elements(acol) eq 0) then acol = 1
  if (n_elements(mcol) eq 0) then mcol = 2                 ; Mass
  if (n_elements(magcol) eq 0) then magcol = [10, 12]      ; V-I vs V
  if (n_elements(am_ab) eq 0) then am_ab = [0.748, 0.358]  ; AV/AB, AI/AB
  if (n_elements(ab) eq 0) then ab = 0.0
  if (n_elements(nab) eq 0) then nab = 10
  if (n_elements(logscl) eq 0) then logscl = 0.05
  if (n_elements(ktype) eq 0) then ktype = 'GAUSS'   ; Kernel for Hess diagr.
  if (n_elements(kwidth) eq 0) then kwidth = 0.01    ; Kernel width 
  if (n_elements(hdim) eq 0) then hdim = 200         ; Dimension of Hess diagr.
  if (n_elements(ymag) eq 0) then ymag=0             ; plot magcol[ymag]
                                                       ; on y-axis
  if (n_elements(mrng) eq 0) then mrng = [0.15, 100]   ; Mass range *for IMF
                                                       ; normalisation*
  if (n_elements(fbin) eq 0) then fbin = 0    ; Fraction of stars that have a
                                              ; companion
  if (n_elements(dbin) eq 0) then dbin = 'DELTA'
  if (n_elements(qmin) eq 0) then qmin = 0.5  ; Minimum mass ratio to consider
  if (n_elements(qmax) eq 0) then qmax = 1.0  ; Maximum mass ratio to consider
  if (n_elements(nqsub) eq 0) then nqsub = 10 ; Number of bins for sampling
                                              ; companion mass function

  data = rdtxt(isofile,[acol, mcol, magcol[0], magcol[1]])

  sda = sort(abs( data(0,*) - logage ))
  loga = data(0,sda[0])
  print, loga
  w = where(data(0,*) eq loga)
  nw = n_elements(w)

  m0 = 1D*transpose(data(1,w))
  mag1_0 = 1D*transpose(data(2,w))
  mag2_0 = 1D*transpose(data(3,w))
  
  mass   = rebin(m0, nsub*nw)
;  mag1   = rebin(mag1_0, nsub*nw)
;  mag2   = rebin(mag2_0, nsub*nw)
  mag1   = interpol(mag1_0, m0, mass)
  mag2   = interpol(mag2_0, m0, mass)

  dm     = dblarr(nw*nsub)
  weight = dblarr(nw*nsub)
  mtot   = 0.

  for i=0,n_elements(mass)-1 do begin
    if (i lt n_elements(mass)-1) then dm[i] = mass[i+1] - mass[i]
    weight[i] = (1 - fbin) * dndm(mass[i])*dm[i]
  end

  ; Add binaries 

  if (fbin gt 0) then begin
    print, 'Adding binaries, fbin=',strn(fbin, format='(F5.2)'), ' dbin=',dbin
    pmass = mass
    bmag1 = dblarr(n_elements(mag1)*nqsub) 
    bmag2 = bmag1
    bweight = bmag1

    k = 0L
    for i=0, n_elements(pmass)-1 do begin
      p = qmin + (qmax-qmin)*dindgen(nqsub)/(nqsub-1) ; Mass ratios 0.5 - 1.0
      fp = dblarr(nqsub)
      for j=0,nqsub-1 do fp[j] = dndp(pmass[i], p[j], dbin=dbin, qmin=qmin, qmax=qmax)
      for j=0,nqsub-1 do begin
	magp1 = interpol(mag1, mass, pmass[i])
	magp2 = interpol(mag2, mass, pmass[i])
	mags1 = interpol(mag1, mass, pmass[i]*p[j])
	mags2 = interpol(mag2, mass, pmass[i]*p[j])

	magc1 = mags1 - 2.5*alog10(1 + 10^(-0.4*(magp1-mags1)))
	magc2 = mags2 - 2.5*alog10(1 + 10^(-0.4*(magp2-mags2)))

	bmag1[k] = magc1
	bmag2[k] = magc2
	bweight[k] = fbin*dndm(pmass[i])*dm[i] * fp[j]/total(fp)
	k = k+1
      end
    end

    mag1 = [mag1, bmag1]
    mag2 = [mag2, bmag2]
    weight = [weight, bweight]
  end

  ; Integrate IMF -
  ; remember to count the mass outside the isochrone range
  ; For a simple power-law IMF the procedure below is a bit clumsy but
  ; we want something that works generally
  ;
  ; Also needs to take into account binaries

  logmrng = alog(mrng)
  ndm = 10000.
  dlogm = (logmrng[1] - logmrng[0]) / (ndm-1)
  mm = exp( logmrng[0] + findgen(ndm) * dlogm)
  dm = mm * dlogm

;  mtot = total(mm*dndm(mm) * dm)
;  print, 'Mtot = ',mtot

  mtot = 0.
  for i=0,n_elements(mm)-1 do begin
    mtot = mtot + (1-fbin)*dndm(mm[i])*dm[i]*mm[i]
    if (fbin gt 0) then begin
      fp = dblarr(nqsub)
      for j=0,nqsub-1 do fp[j] = dndp(mm[i], p[j], dbin=dbin, qmin=qmin, qmax=qmax)
      mtot =  mtot + $
        fbin*dndm(mm[i])*dm[i]* (mm[i]+total(fp*p*mm[i])/total(fp))
    end
  end
;  print, 'Mtot = ',mtot

  ; Generate the Hess diagrams, either for a single extinction value
  ; or a range (by simulating several discrete extinctions and averaging)

  if (ymag eq 0) then magy = mag1 else magy=mag2

;  print, n_elements(mag1), n_elements(mag2), n_elements(weight)

  if (n_elements(ab) eq 1) then $
    hess_sfh, mag1+ab*am_ab[0], mag2+ab*am_ab[1], $
      weights=weight, xr=xr, yr=yr, logscl=0.05, $
      errm1=errm1, errm2=errm2, $
      pixels=pixels, xp=xp, yp=yp, ymag=ymag, /noshow, $
      nx=hdim, ny=hdim, $
      cpixels=cpixels, cmode=cmode, ktype=ktype, sigker=kwidth $
  else begin
    for i=0,nab-1 do begin
      abi = ab[0] + (ab[1]-ab[0])*(1.0*i)/(nab-1)
      hess_sfh, mag1+abi*am_ab[0], mag2+abi*am_ab[1], $
        weights=weight, xr=xr, yr=yr, logscl=0.05, $
	errm1=errm1, errm2=errm2, $
        pixels=pixels, xp=xp, yp=yp, ymag=ymag, /noshow, $
        nx=hdim, ny=hdim, $
	cpixels=cpixels, cmode=cmode, ktype=ktype, sigker=kwidth
      if (i eq 0) then psum = pixels else psum = psum + pixels
    end
    pixels = psum / nab
  end

; Normalise to a total mass of 1 Msun over MRNG

  pixels = pixels / mtot

; Plot the Hess diagram unless otherwise noted

  if (n_elements(noshow) eq 0) then $
    phess, pixels, xr, yr, logscl=logscl, neg=neg, charsize=charsize, $
      xtickformat=xtickformat, ytickformat=ytickformat, xtitle=xtitle, $
      ytitle=ytitle, title=title
;    plot, xr, yr, xr=xr, yr=yr, xstyle=1, ystyle=1, /nodata
;    pp = alog10(pixels/max(pixels)+0.05)
;    pscl = (!D.TABLE_SIZE-1)*(pp-min(pp))/(max(pp)-min(pp))
;    tvimage, pscl, /overplot
;
;    axis, xaxis=0, xstyle=1, xtickformat='(A1)', color=color
;    axis, xaxis=1, xstyle=1, xtickformat='(A1)', color=color
;    axis, yaxis=0, ystyle=1, ytickformat='(A1)', color=color
;    axis, yaxis=1, ystyle=1, ytickformat='(A1)', color=color
;  end


;  oplot, mag1-mag2, mag2, psym=3, color=fsc_color('red')

end
